#include "libopgp.h"
#include <md5.h>
#include <sha.h>
#include <ripemd.h>
#include <md2.h>
#include "haval.h"
#include "tiger.h"

#ifdef NO_HAVAL
#define HAVAL_Init NULL
#define HAVAL_Update NULL
#define HAVAL_Final NULL
static char havalContext[0];
#else
static void HAVAL_Init(void *x)
{
  havalInit(x, 3, 160);
}
static void HAVAL_Update(void *x, void *y, int z)
{
  havalUpdate(x, y, z);
}
static void HAVAL_Final(void *x, void *y)
{
  havalFinal(y, x);
}

#endif

#ifdef NO_TIGER
#define TIGER_Init NULL
#define TIGER_Update NULL
#define TIGER_Final NULL
static char TIGER_CTX[0];
#endif

#define MAXHASH 7

const u32_t PGP_hctxsiz[MAXHASH] =
{-1, sizeof(MD5_CTX), sizeof(SHA_CTX), sizeof(RIPEMD160_CTX),
 sizeof(havalContext), sizeof(MD2_CTX), sizeof(TIGER_CTX)};

const char PGP_hbnr[MAXHASH][24] =
{"SHA1", "RIPEMD160", "HAVAL-5-160", "MD2", "TIGER192"};

void *PGP_hini(int ha)
{
  u8_t *ctx;
  int siz;

  if (ha >= MAXHASH)
    return NULL;
  if (0 >= (siz = PGP_hctxsiz[ha]))
    return NULL;

  ctx = malloc(siz + 2);

  *ctx++ = ha;
  *ctx++ = 0;                   /* word align */

  if (ha == 2)
    SHA1_Init(ctx);
  else if (ha == 1)
    MD5_Init(ctx);
  else if (ha == 3)
    RIPEMD160_Init(ctx);
  else if (ha == 5)
    MD2_Init(ctx);
  else
    ErrDisplay("BadHashAlg");

  ctx--;
  return --ctx;
}

void PGP_hblk(u8_t * ctx, u8_t * buf, u32_t len)
{
  int ha = *ctx++;
  if (ha >= MAXHASH)
    exit(-1);

  if (ha == 2)
    SHA1_Update(++ctx, buf, len);
  else if (ha == 1)
    MD5_Update(++ctx, buf, len);
  else if (ha == 3)
    RIPEMD160_Update(++ctx, buf, len);
  else if (ha == 5)
    MD2_Update(++ctx, buf, len);

}

void PGP_hfin(u8_t * buf, u8_t * ctx)
{
  int ha = *ctx++;
  if (ha >= MAXHASH)
    exit(-1);

  if (ha == 2)
    SHA1_Final(buf, ++ctx);
  else if (ha == 1)
    MD5_Final(buf, ++ctx);
  else if (ha == 3)
    RIPEMD160_Final(buf, ++ctx);
  else if (ha == 5)
    MD2_Final(buf, ++ctx);
  ctx--;

  free(--ctx);

}

u8_t PGP_hlen[MAXHASH] =
{0, 16, 20, 20, 20, 16, 18};

static const u8_t PGP_hpfx[MAXHASH][24] =
{
  {0},
  {0x30, 0x20, 0x30, 0x0c, 0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
   0x02, 0x05, 0x05, 0x00, 0x04, 0x10,},
  {0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x1a, 0x05,
   0x00, 0x04, 0x14,},
  {0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03, 0x02, 0x01, 0x05,
   0x00, 0x04, 0x14,},
  {0x30, 0x27, 0x30, 0x0f, 0x04, 0x0b, 0x48, 0x41, 0x56, 0x41, 0x4c, 0x2d,
   0x35, 0x2d, 0x31, 0x36, 0x30, 0x05, 0x00, 0x04, 0x14,} /* FAKE */ ,
  {0x30, 0x20, 0x30, 0x0c, 0x06, 0x08, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d,
   0x02, 0x02, 0x05, 0x00, 0x04, 0x10,},
  {0x30, 0x28, 0x30, 0x0c, 0x04, 0x08, 0x54, 0x49, 0x47, 0x45, 0x52, 0x31,
   0x39, 0x32, 0x05, 0x00, 0x04, 0x18,},  /* FAKE */
};
static const u8_t PGP_hpfxlen[MAXHASH] =
{0, 18, 15, 15, 21 - 21, 18, 18 - 18};

int PGP_hder(u8_t halg, u8_t * hbuf)
{
  int k = PGP_hpfxlen[halg];
  if (k) {
    memmove(&hbuf[k], hbuf, PGP_hlen[halg]);
    memcpy(hbuf, PGP_hpfx[halg], k);
  }
  return k + PGP_hlen[halg];
}

/*--------------------------------------------------*/
/* string to key */
void PGP_khash(u8_t ha, u32_t hbytes, u32_t keymat, u32_t keysize,
               u8_t * hbuf, u8_t * hout)
{
  u32_t loops;
  int block;
  u8_t *hctx;

  for (block = 0; block * PGP_hlen[ha] < keysize; block++) {
    hctx = PGP_hini(ha);

    memset(hout, 0, block);     /* pass prehashes pass-1 zeros */
    if (block)
      PGP_hblk(hctx, hout, block);
    loops = hbytes;             /* loops over whole text */
    while (loops > keymat)
      PGP_hblk(hctx, hbuf, keymat), loops -= keymat;
    PGP_hblk(hctx, hbuf, loops);

    PGP_hfin(hout, hctx);
    hout += PGP_hlen[ha];
  }
}
